"""Module with the definition of lead liquid metal object class,
i.e., Lead"""
import copy
from ._constants import LEAD_MELTING_TEMPERATURE
from ._constants import LEAD_MELTING_LATENT_HEAT, LEAD_BOILING_TEMPERATURE
from ._constants import SOBOLEV_KEYWORD, LEAD_VAPORISATION_HEAT, P_ATM
from ._lbh15 import LiquidMetalInterface


class Lead(LiquidMetalInterface):
    """
    Class to model liquid lead properties at a given temperature

    Parameters
    ----------
    p : float, optional
        Pressure in [Pa], by default atmospheric pressure, i.e.,
        101325.0 Pa
    \\**kwargs : dict
        Dictionary that specifies the quantity from which the object shall
        be initialized. The default available ones are:

        - **T** (float) : temperature [K]
        - **p_s** (float) : saturation vapour pressure [Pa]
        - **sigma** (float) : surface tension [N/m]
        - **rho** (float) : density [Kg/m^3]
        - **alpha** (float) : thermal expansion coefficient [1/K]
        - **u_s** (float) : speed of sound [m/s]
        - **beta_s** (float) : isentropic compressibility [1/Pa]
        - **cp** (float) : specific heat capacity [J/(kg*K)]
        - **h** (float) : specific enthalpy \
        (in respect to melting point) [J/kg]
        - **mu** (float) : dynamic viscosity [Pa*s]
        - **r** (float) : electrical resistivity [Ohm*m]
        - **k** (float) : thermal conductivity [W/(m*K)]

    Examples
    --------
    Compare :class:`.lead.Lead` specific heat values at T=800 K
    with with using 'sobolev2011' and 'gurvich1991':

    >>> liquid_lead_1 = Lead(T=800)  # 'sobolev2011'
    >>> liquid_lead_1.cp
    144.31634999999997
    >>> Lead.set_correlation_to_use('cp', 'gurvich1991')
    >>> liquid_lead_2 = Lead(T=800)
    >>> liquid_lead_2.cp
    144.66006199999998
    """
    _default_corr_to_use = {'cp': SOBOLEV_KEYWORD}
    _correlations_to_use = copy.deepcopy(_default_corr_to_use)
    _roots_to_use = {'cp': 0}
    _properties_module = 'lbh15.properties.lead_properties'

    def __init__(self, p=P_ATM, **kwargs):
        self._guess = LEAD_MELTING_TEMPERATURE*1.7
        super().__init__(p=p, **kwargs)

    def _set_constants(self):
        """
        Sets the class constants
        """
        self._T_m0 = LEAD_MELTING_TEMPERATURE
        self._Q_m0 = LEAD_MELTING_LATENT_HEAT
        self._T_b0 = LEAD_BOILING_TEMPERATURE
        self._Q_b0 = LEAD_VAPORISATION_HEAT



"""Module with the definition of thermophysical property objects for bismuth"""
from .interface import PropertyInterface, range_warning
from .._constants import BISMUTH_MELTING_TEMPERATURE as T_m0
from .._constants import BISMUTH_BOILING_TEMPERATURE as T_b0
from .._constants import SOBOLEV_KEYWORD, P_ATM


class p_s(PropertyInterface):
    """
    Liquid bismuth saturation vapour pressure
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute saturation vapour pressure

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        saturation vapour pressure in [Pa] : float
        """
        import numpy as np
        return 2.67e10 * np.exp(-22858/T)

    def initialization_helper(self, property_value):
        """
        Returns a temperature guess according to the value
        of the saturation vapour pressure

        Parameters
        ----------
        property_value : float
            saturation vapour pressure in [Pa]
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        rvalue : float
            Temperature guess in [K]
        """
        if property_value < 1e-2:
            rvalue = 800
        elif 1e-2 <= property_value < 1e2:
            rvalue = 1200
        else:
            rvalue = 2000

        return rvalue

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return SOBOLEV_KEYWORD

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, T_b0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[Pa]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "saturation vapour pressure"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"


class sigma(PropertyInterface):
    """
    Liquid bismuth surface tension
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute surface tension

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        surface tension in [N/m] : float
        """
        return (420.8 - 0.081*T)*1e-3

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return SOBOLEV_KEYWORD

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, 1400.0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[N/m]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "surface tension"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"


class rho(PropertyInterface):
    """
    Liquid bismuth density
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute density

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        density in [kg/m^3] : float
        """
        rho_0 = 10725 - 1.22*T
        u_s_val = u_s().correlation(T, p)
        cp_val = cp().correlation(T, p)
        alpha_val = alpha().correlation(T, p)
        return rho_0 + ((u_s_val**-2 + T*alpha_val**2/cp_val) * (p - P_ATM))

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return "imbeni1998"

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, T_b0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[kg/m^3]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "density"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"


class alpha(PropertyInterface):
    """
    Liquid bismuth thermal expansion coefficient
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute thermal expansion coefficient

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        thermal expansion coefficient in [1/K] : float
        """
        return 1/(8791 - T)

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, T_b0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[1/K]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "thermal expansion coefficient"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"


class u_s(PropertyInterface):
    """
    Liquid bismuth sound velocity
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute sound velocity

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        sound velocity in [m/s] : float
        """
        return 1616 + 0.187*T - 2.2e-4*T**2

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return SOBOLEV_KEYWORD

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, 1800.0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[m/s]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "sound velocity"

    @property
    def description(self):
        """
        str : property description
        """
        return "Sound velocity in liquid bismuth"


class beta_s(PropertyInterface):
    """
    Liquid bismuth isentropic compressibility
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute isentropic compressibility

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        isentropic compressibility in [1/Pa] : float
        """
        rho_obj = rho()
        u_s_obj = u_s()
        return 1/(rho_obj.correlation(T, p) * u_s_obj.correlation(T, p)**2)

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, 1800.0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[1/Pa]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "isentropic compressibility"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"


class cp(PropertyInterface):
    """
    Liquid bismuth specific heat capacity
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute specific heat capacity

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        specific heat capacity in [J/(kg*K)] : float
        """
        return 118.2 + 5.934e-3*T + 7.183e6*T**-2

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return "imbeni1998"

    @property
    def is_injective(self):
        """
        bool : True if correlation is injective,
        False otherwise
        """
        return False

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, T_b0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[J/(kg*K)]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "specific heat capacity"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"


class h(PropertyInterface):
    """
    Liquid bismuth specific enthalpy
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute specific enthalpy

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        specific enthalpy in [J/kg] : float
        """
        return (118.2*(T - T_m0)
                + 2.967e-3*(T**2 - T_m0**2)
                - 7.183e6*(T**-1 - T_m0**-1))

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return SOBOLEV_KEYWORD

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, T_b0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[J/kg]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "specific enthalpy"

    @property
    def description(self):
        """
        str : property description
        """
        return (f"Liquid bismuth {self.long_name} "
                "(as difference with respect to "
                "the melting point enthalpy)")


class mu(PropertyInterface):
    """
    Liquid bismuth dynamic viscosity
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute dynamic viscosity

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        dynamic viscosity in [Pa*s] : float
        """
        import numpy as np
        return 4.456e-4*np.exp(780/T)

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return "lucas1984b"

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, 1300.0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[Pa*s]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "dynamic viscosity"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"


class r(PropertyInterface):
    """
    Liquid bismuth electrical resistivity
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute electrical resistivity

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        electrical resistivity in [Ohm*m] : float
        """
        return (98.96 + 0.0554*T)*1e-8

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [545.0, 1423.0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[Ohm*m]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "electrical resistivity"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"


class k(PropertyInterface):
    """
    Liquid bismuth thermal conductivity
    property class
    """
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Correlation used to compute thermal conductivity

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False

        Returns
        -------
        thermal conductivity in [W/(m*K)] : float
        """
        return 7.34 + 9.5e-3*T

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return "touloukian1970b"

    @property
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        return [T_m0, 1000.0]

    @property
    def units(self):
        """
        str : property units
        """
        return "[W/(m*K)]"

    @property
    def long_name(self):
        """
        str : property long name
        """
        return "thermal conductivity"

    @property
    def description(self):
        """
        str : property description
        """
        return f"Liquid bismuth {self.long_name}"



"""Module with the definition of liquid metal object base class,
i.e., LiquidMetalInterface"""
import warnings
import sys
import inspect
import importlib
import platform
import copy
from abc import ABC, abstractmethod
from .properties.interface import PropertyInterface
from ._constants import P_ATM


class LiquidMetalInterface(ABC):
    """
    Abstract class that defines liquid metal properties object

    Parameters
    ----------
    p : float, optional
        Pressure in [Pa], by default atmospheric pressure, i.e.,
        101325.0 Pa
    **kwargs : dict
        Dictionary that specifies the quantity from which the parameter shall
        be initialized. The default available ones are:

        - 'T' : temperature [K]
        - 'p_s' : saturation vapour pressure [Pa]
        - 'sigma' : surface tension [N/m]
        - 'rho' : density [Kg/m^3]
        - 'alpha' : thermal expansion coefficient [1/K]
        - 'u_s': speed of sound [m/s]
        - 'beta_s' : isentropic compressibility [1/Pa]
        - 'cp' : specific heat capacity [J/(kg*K)]
        - 'h' : specific enthalpy (in respect to melting point) [J/kg]
        - 'mu' : dynamic viscosity [Pa*s]
        - 'r' : electrical resistivity [Ohm*m]
        - 'k' : thermal conductivity [W/(m*K)]
    """
    _T_m0 = 0
    _Q_m0 = 0
    _T_b0 = 0
    _Q_b0 = 0
    _guess = 0
    _correlations_to_use = {}
    _roots_to_use = {}
    _default_corr_to_use = {}
    _properties_module = ""
    __p = 0
    __T = 0
    __custom_properties_path = {}

    def __init__(self, p=P_ATM, **kwargs):
        self.__assign_p(p)
        self.__properties = {}
        self.__corr2use = copy.deepcopy(self.__class__._correlations_to_use)
        self.__fill_instance_properties()
        self.__fill_class_attributes(kwargs)

    @property
    def T_m0(self):
        """
        float : melting temperature [K]
        """
        return self._T_m0

    @property
    def Q_m0(self):
        """
        float : melting latent heat [J/kg]
        """
        return self._Q_m0

    @property
    def T_b0(self):
        """
        float : boiling temperature [K]
        """
        return self._T_b0

    @property
    def Q_b0(self):
        """
        float : vaporisation heat [J/kg]
        """
        return self._Q_b0

    @property
    def p(self):
        """
        float : pressure adopted for property calculation, i.e.,
        atmospheric pressure
        """
        return self.__p

    @p.setter
    def p(self, p):
        """
        Set liquid metal pressure

        Parameters
        ----------
        p : float
            Pressure in [Pa]
        """
        self.__assign_p(p)

    @property
    def T(self):
        """
        float : temperature used to compute properties [K]
        """
        return self.__T

    @T.setter
    def T(self, T):
        """
        Set liquid metal temperature

        Parameters
        ----------
        T : float
            Temperature in [K]
        """
        self.__assign_T(T)

    @property
    def Pr(self):
        """
        float : Prandtl number [-]
        """
        return self.cp * self.mu / self.k

    @property
    def correlations_used(self):
        """
        Returns the dictionary with the specific
        correlation used for each specified property

        Returns
        -------
        dict
        """
        return copy.deepcopy(self.__corr2use)

    def change_correlation_to_use(self, property_name, correlation_name):
        """
        Changes property correlation, if property is defined as instance
        attribute.

        Parameters
        ----------
        property_name : str
            Name of the thermophysical property
        correlation_name : str
            Name of the correlation
        """
        key = self.__generate_key(property_name)
        if (key in self.__properties and
                self.__properties[key].correlation_name != correlation_name):
            self.__corr2use[property_name] = correlation_name
            self.__fill_instance_properties()

    def check_temperature(self, T):
        """
        Checks that the provided temperature
        belongs to the correct temperature range, i.e.,
        liquid temperature range.

        Parameters
        ----------
        T : float
            Temperature in [K]

        Returns
        -------
        rvalue : bool
            True if check is ok, False otherwise
        error_message : str
            Contains the error message (if any) associated
            to the temperature check
        """
        if self.T_m0 < T < self.T_b0:
            rvalue = True
            error_message = ""
        else:
            rvalue = False
            if T >= self.T_b0:
                error_message = ("Temperature must be smaller than "
                                 f"boiling temperature ({self.T_b0:.2f} [K]), "
                                 f"{T:.2f} [K] was provided")
            elif 0 < T <= self.T_m0:
                error_message = ("Temperature must be larger than "
                                 f"melting temperature ({self.T_m0:.2f} [K]), "
                                 f"{T:.2f} [K] was provided")
            else:
                error_message = ("Temperature must be "
                                 "strictly positive, "
                                 f"{T:.2f} [K] was provided")

        return rvalue, error_message

    @classmethod
    def properties_for_initialization(cls):
        """
        List of available properties that can be used for
        initialization

        Returns
        -------
        list
        """
        obj_list = cls.__load_properties()
        obj_list += cls.__load_custom_properties()
        rvalue = ['T'] + [obj_list[i].name for i in range(len(obj_list))]
        return list(dict.fromkeys(rvalue))

    @classmethod
    def correlations_available(cls):
        """
        Dictionary of correlations available for each property

        Returns
        -------
        dict
        """
        obj_list = cls.__load_properties()
        obj_list += cls.__load_custom_properties()
        rvalue = {}
        for obj in obj_list:
            corr_name = obj.correlation_name
            if obj.name in rvalue:
                if isinstance(rvalue[obj.name], list):
                    rvalue[obj.name].append(corr_name)
                else:
                    rvalue[obj.name] = [rvalue[obj.name]] + [corr_name]
            else:
                rvalue[obj.name] = corr_name
        return rvalue

    @classmethod
    def set_correlation_to_use(cls, property_name, correlation_name):
        """
        Set specific correlation to use for property

        Parameters
        ----------
        property_name : str
            Name of the property
        correlation_name : str
            Name of the correlation
        """
        cls._correlations_to_use[property_name] = correlation_name

    @classmethod
    def set_root_to_use(cls, property_name, root_index):
        """
        Set which temperature root shall be used for initialization
        from property. Temperature roots are sorted in ascending order, i.e,
        T_i <= T_j with i < j. Used only if property correlation
        is not injective

        Parameters
        ----------
        property_name : str
            Name of the property
        root_index : int
            Index used to choose the temperature root
        """
        cls._roots_to_use[property_name] = root_index

    @classmethod
    def set_custom_properties_path(cls, file_path):
        """
        Set absolute path of file where looking for custom properties

        Parameters
        ----------
        file_path : str
            absolute path of file where custom properties are implemented
        """
        char = '/' if 'Linux' in platform.system() else '\\'
        res = file_path.split(char)
        file_name = res[-1][:-3]
        path = file_path[:-len(res[-1])]
        lm_name = cls.__name__
        if lm_name not in cls.__custom_properties_path:
            cls.__custom_properties_path[lm_name] = {}
        cls.__custom_properties_path[lm_name][path] = file_name

    @classmethod
    def correlations_to_use(cls):
        """
        Returns the dictionary with the specific
        correlation to use for each specified property

        Returns
        -------
        dict
        """
        return copy.deepcopy(cls._correlations_to_use)

    @classmethod
    def roots_to_use(cls):
        """
        Returns the dictionary with temperature
        roots to use for each specified property

        Returns
        -------
        dict
        """
        return copy.deepcopy(cls._roots_to_use)

    def __compute_T(self, input_value, input_property):
        """
        Computes the temperature in [K] that is then set as value
        of the object property

        Parameters
        ----------
        input_value : float
            value of the property used to compute the temperature
        input_property : str
            name of the property used to perform the calculation
        """
        if input_property == 'T':
            rvalue = input_value
        else:
            function_of_T = None
            helper = None
            is_injective = False
            key = self.__generate_key(input_property)
            if key in self.__properties:
                function_of_T = self.__properties[key].correlation
                helper = self.__properties[key].initialization_helper
                is_injective = self.__properties[key].is_injective

            if function_of_T is not None:
                from scipy.optimize import fsolve

                if helper(input_value) is not None:
                    self._guess = helper(input_value)

                def function_to_solve(T, target):
                    return function_of_T(T, self.__p) - target

                if is_injective:
                    res = fsolve(function_to_solve, x0=[self._guess],
                                 args=(input_value), xtol=1e-10)
                    rvalue = res[0]
                else:
                    index = (self._roots_to_use[input_property]
                             if input_property in self._roots_to_use
                             else 0)
                    res = fsolve(function_to_solve,
                                 x0=[self._guess, 3*self._guess],
                                 args=(input_value), xtol=1e-10)
                    if len(res) > index - 1:
                        rvalue = res[index]
                    else:
                        rvalue = res[0]

        return rvalue

    def __fill_instance_properties(self):
        """
        Fills instance properties.
        """
        property_obj_list = self.__load_properties()
        property_obj_list += self.__load_custom_properties()
        for property_object in property_obj_list:
            # always add property if specific correlation is not specified
            name = property_object.name
            add_property = (not self.__corr2use or
                            name not in self.__corr2use.keys())
            # if correlation was specified, check that correlation names match
            if not add_property:
                add_property = (property_object.correlation_name ==
                                self.__corr2use[name])
            if add_property:
                self.__add_property(property_object)

        keys_to_remove = self.__check_properties()
        for key_to_remove in keys_to_remove:
            self.__corr2use.pop(key_to_remove)
            if key_to_remove in self._correlations_to_use:
                self._correlations_to_use.pop(key_to_remove)

    def __fill_class_attributes(self, kwargs):
        """
        Fills all the class attributes.

        Parameters
        ----------
        T : float
            Temperature in [K]
        """
        if len(kwargs) != 1:
            raise ValueError("One and only one property at "
                             "time can be used for initialization. "
                             f"{len(kwargs)} were provided")

        valid_prop = self.properties_for_initialization()
        input_property = list(kwargs.keys())[0]
        input_value = kwargs[input_property]
        if input_property not in valid_prop:
            list_to_print = "\n\n"
            for sym in valid_prop:
                list_to_print += sym+"\n"
            list_to_print += "\n"
            raise ValueError("Initialization can be done only with one of "
                             f"the following properties:{list_to_print}"
                             f"{input_property} was provided")

        self._set_constants()
        temperature = self.__compute_T(input_value, input_property)
        self.__assign_T(temperature)

    def __assign_T(self, T):
        """
        Function used to set temperature, checking if temperature value
        is valid, i.e., strictly positive and (if yes) inside
        the range of melting and boiling temperature.

        Parameters
        ----------
        T : float
            Temperature in [K]
        """
        temp_ok, error_message = self.check_temperature(T)
        if temp_ok:
            self.__T = T
        else:
            raise ValueError(error_message)

    def __assign_p(self, p):
        """
        Function used to set pressure, checking if
        pressure value is valid, i.e., strictly positive.

        Parameters
        ----------
            p : float
            Pressure in [Pa]
        """
        if p > 0:
            self.__p = p
        else:
            raise ValueError("Pressure must be "
                             "strictly positive, "
                             f"{p:.2f} [Pa] was provided")

    def __add_property(self, property_object):
        """
        Adds the property to class attributes. In particular, it adds
        '<prpertyObject.name>_info' as class method to get additional
        information on property. Moreover it adds property_object
        to instance dictionary which will be used in dunder __getattr__
        to return attribute '<prpertyObject.name>'.

        Parameters
        ----------
        property_object : :class:`_properties.PropertyInterface`
            Object which inherits from :class:`_properties.PropertyInterface`
        """
        key = self.__generate_key(property_object.name)
        self.__properties[key] = property_object

        def new_property_info(print_info=True, n_tab=0):
            return self.__properties[key].info(self.__T, self.__p,
                                               print_info, n_tab)

        setattr(self, property_object.name+"_info",
                new_property_info)

    def __generate_key(self, property_name):
        """
        Generates the key that will be used to store the property
        in class dictionary

        Parameters
        ----------
        property_name : str
            Name of the property

        Returns
        -------
        str
            Generated key
        """
        return property_name + '_' + type(self).__name__

    def __check_properties(self):
        keys_to_remove = []
        update_properties = False
        for key in self.__corr2use.keys():
            prop_key = self.__generate_key(key)
            corr_name = self.__corr2use[key]
            is_in_default = key in self._default_corr_to_use
            remove_property = False

            if prop_key not in self.__properties:
                if not is_in_default:
                    warnings.warn(f"Could not find property '{key}' "
                                  f"implementing '{corr_name}' correlation. "
                                  "\nGoing to restore package default one, "
                                  "if any.",
                                  stacklevel=5)
                    remove_property = True
                    corr_avail = self.correlations_available()
                    if key in corr_avail:
                        isstr = isinstance(corr_avail[key], str)
                        self.__corr2use[key] = (corr_avail[key] if isstr
                                                else corr_avail[key][-1])
                        update_properties = True
                else:
                    def_corr_name = self._default_corr_to_use[key]
                    warnings.warn(f"Could not find property '{key}' "
                                  f"implementing '{corr_name}' correlation. "
                                  "\nGoing to restore default correlation "
                                  f"'{def_corr_name}'.",
                                  stacklevel=5)
                    self.__corr2use[key] = def_corr_name
                    remove_property = True
                    update_properties = True
            else:
                if corr_name != self.__properties[prop_key].correlation_name:
                    warnings.warn(f"Could not find property '{key}' "
                                  f"implementing '{corr_name}' correlation. "
                                  "\nGoing to remove it from correlations "
                                  "to use.",
                                  stacklevel=5)
                    remove_property = not is_in_default
                    if not remove_property:
                        self.__corr2use[key] = self._default_corr_to_use[key]

            if remove_property:
                keys_to_remove.append(key)

        if update_properties:
            self.__fill_instance_properties()

        return keys_to_remove

    @classmethod
    def __load_properties(cls):
        """
        Loads property objects corresponding to liquid metal

        Returns
        -------
        list
            list of property objects, i.e. of classes which inherit from
            :class:`_properties.PropertyInterface`
        """
        property_obj_list = []
        module = cls._properties_module
        if module:
            for _, obj in inspect.getmembers(sys.modules[module]):
                if inspect.isclass(obj) and obj is not PropertyInterface:
                    if issubclass(obj, PropertyInterface):
                        property_obj_list.append(obj())
        return property_obj_list

    @classmethod
    def __load_custom_properties(cls):
        """
        Load custom property objects

        Returns
        -------
        list
            list of property objects, i.e. of classes which inherit from
            :class:`_properties.PropertyInterface`
        """
        customproperty_obj_list = []
        if cls.__name__ in cls.__custom_properties_path:
            lm_path = cls.__custom_properties_path[cls.__name__]
            for path in lm_path.keys():
                if path not in sys.path:
                    sys.path.append(path)
                module_name = lm_path[path]
                module = importlib.import_module(module_name)
                for _, obj in inspect.getmembers(module):
                    if inspect.isclass(obj) and obj is not PropertyInterface:
                        if issubclass(obj, PropertyInterface):
                            customproperty_obj_list.append(obj())
        return customproperty_obj_list

    @abstractmethod
    def _set_constants(self):
        """
        Sets the class constants
        """
        raise NotImplementedError(f"{type(self).__name__}._set_constants "
                                  "NOT IMPLEMENTED")

    def __getattr__(self, name):
        key = self.__generate_key(name)
        if key not in self.__properties:
            raise AttributeError(f"'{type(self).__name__}' object "
                                 f"has no attribute '{name}'")

        return self.__properties[key].correlation(self.__T, self.__p, True)

    def __str__(self):
        rvalue = (f"{type(self).__name__} liquid metal "
                  f"@(T={self.T:.2f} [K], p={self.p:.2f} [Pa])\n")
        rvalue += "\nConstants:\n"
        rvalue += f"\tMelting Temperature: {self.T_m0:.2f} [K]\n"
        rvalue += f"\tBoiling Temperature: {self.T_b0:.2f} [K]\n"
        rvalue += f"\tMelting latent heat: {self.Q_m0:.2f} [J/kg]\n"
        rvalue += f"\tVaporisation heat: {self.Q_b0:.2f} [J/kg]\n"
        rvalue += "\nThermophysical properties:\n"
        for key in self.__properties:
            prop_name = key.replace("_"+type(self).__name__, "")
            info = getattr(self, prop_name+"_info")(print_info=False, n_tab=1)
            rvalue += info + "\n"
        return rvalue

    def __repr__(self):
        rvalue = f"{type(self).__name__}(T={self.T:.2f}, "
        for key in self.__properties:
            property_name = key.replace("_"+type(self).__name__, "")
            attr_value = getattr(self, property_name)
            if attr_value < 1e-2:
                rvalue += f"{property_name}={attr_value:.2e}, "
            else:
                rvalue += f"{property_name}={attr_value:.2f}, "
        rvalue = rvalue[:-2]
        rvalue += ")"
        return rvalue



"""Module with the definition of thermophysical property object base class,
i.e., PropertyInterface. Definition of decorator
for validity range check as well."""
import warnings
from abc import ABC, abstractmethod
from .._constants import P_ATM


def range_warning(function):
    """
    Decorator used to check validity range
    of correlation
    """
    def wrapper(*args):
        range_lim = args[0].range
        p_name = args[0].long_name
        temp = args[1]
        if hasattr(temp, "__len__"):
            temp = temp[0]
        if temp < range_lim[0] or temp > range_lim[1]:
            if len(args) == 4:
                warnings.warn(f"The {p_name} is requested at "
                              f"temperature value of {temp:.2f} K "
                              "that is not in validity range "
                              f"[{range_lim[0]:.2f}, {range_lim[1]:.2f}] K",
                              stacklevel=3)
        return function(*args)
    return wrapper


class PropertyInterface(ABC):
    """
    Abstract class that defines thermo-physical property interface.
    Derived classes must override:

        - :func:`~PropertyInterface.correlation`
        - :attr:`~.PropertyInterface.range`
        - :attr:`~.PropertyInterface.units`
        - :attr:`~.PropertyInterface.long_name`
        - :attr:`~.PropertyInterface.description`

    Instead, it is not mandatory to override the following attributes:

        - :func:`~PropertyInterface.initialization_helper`, override this \
          method if function's roots are particularly 'difficult' to find \
          (see \
          :func:`lbh15.properties.lead_properties.p_s.initialization_helper`).\
           If not overridden it will be ignored.
        - :attr:`~.PropertyInterface.name`, override this member to use \
          a custom name for the property, otherwise the class name will be used
        - :attr:`~.PropertyInterface.correlation_name` override this member \
          to set a correlation name different from default one
        - :attr:`~.PropertyInterface.is_injective` override this member \
          if the function is not injective, othwerise it will \
          be considered injective
    """
    def __init__(self):
        from numpy import nan
        self.__min = -nan
        self.__max = nan
        self.__T_at_min = -nan
        self.__T_at_max = nan

    def compute_bounds(self):
        """
        Computes the bound of property in validity range, i.e.,
        the minimum and the maximum of the correlation inside
        the validity range, together with the corresponding temperature.
        If this method is not invoked their default value is
        -nan and nan respectively.
        Values are computed using :func:`scipy.optimize.minimize_scalar`
        with "Bounded" solver (for more details please refer to scipy
        documentation)
        """
        from scipy.optimize import minimize_scalar
        min_vals = minimize_scalar(self.correlation,
                                   bounds=self.range,
                                   method="Bounded")
        self.__min = min_vals.fun
        self.__T_at_min = min_vals.x
        if self.__T_at_min - self.range[0] < 5e-4:
            self.__T_at_min = self.range[0]
            self.__min = self.correlation(self.__T_at_min)

        def corr_reciprocal(T):
            return 1/self.correlation(T)

        max_vals = minimize_scalar(corr_reciprocal,
                                   bounds=self.range,
                                   method="Bounded")
        self.__max = self.correlation(max_vals.x)
        self.__T_at_max = max_vals.x
        if self.range[1] - self.__T_at_max < 5e-4:
            self.__T_at_max = self.range[1]
            self.__max = self.correlation(self.__T_at_max)

    def initialization_helper(self, property_value):
        """
        Returns a temperature guess according to the value
        of the property. Used by root finder algorithm if
        return type is not None.

        Parameters
        ----------
        property_value : float
            value of the property

        Returns
        -------
        None
        """
        return None

    def info(self, T, p=P_ATM, print_info=True, n_tab=0):
        """
        Method used to print information about the property
        and the correlation used to compute its value.

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default P_ATM, i.e., 101325.0 Pa
        print_info : bool, optional
            True to print to console, False for getting the
            string, by default True
        n_tab : int, optional
            Number of tabs used to format the information, by default 0

        Returns
        -------
        str, None
            Returns None id print_info=True, otherwise returns the content
            of the information.
        """
        name = self.name
        property_val = self.correlation(T, p)
        if property_val < 1e-2:
            value = f"Value: {property_val:.2e} {self.units}"
        else:
            value = f"Value: {property_val:.2f} {self.units}"
        validity = ("Validity range: "
                    f"[{self.range[0]:.2f}, {self.range[1]:.2f}] K")
        corr_name = f"Correlation name: '{self.correlation_name}'"
        long_name = f"Long name: {self.long_name}"
        units = f"Units: {self.units}"
        description = ("Description:\n{:s}{:s}"
                       .format((n_tab+2)*"\t", self.description))

        all_info = "{:s}{:s}:\n".format(n_tab*"\t", name)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", value)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", validity)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", corr_name)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", long_name)
        all_info += "{:s}{:s}\n".format((n_tab+1)*"\t", units)
        all_info += "{:s}{:s}".format((n_tab+1)*"\t", description)

        rvalue = None
        if print_info:
            print(all_info)
        else:
            rvalue = all_info
        return rvalue

    @abstractmethod
    @range_warning
    def correlation(self, T, p=P_ATM, verbose=False):
        """
        Function that implements the property correlation

        Parameters
        ----------
        T : float
            Temperature in [K]
        p : float, optional
            Pressure in [Pa], by default atmospheric pressure, i.e.,
            101325.0 Pa
        verbose : bool, optional
            True to tell decorator to print warning about
            range check failing, False otherwise. By default False
        """
        raise NotImplementedError(f"{type(self).__name__}.correlation "
                                  "NOT IMPLEMENTED")

    @property
    def name(self):
        """
        str : name of the property
        """
        return type(self).__name__

    @property
    def correlation_name(self):
        """
        str : name of the correlation
        """
        return "lbh15"

    @property
    def is_injective(self):
        """
        bool : True if correlation is injective,
        False otherwise
        """
        return True

    @property
    def min(self):
        """
        float : minimum of property correlation in validity
        range
        """
        return self.__min

    @property
    def max(self):
        """
        float : maximum of property correlation in validity
        range
        """
        return self.__max

    @property
    def T_at_min(self):
        """
        float : temperature corresponding to the minimum of
        property correlation in validity range
        """
        return self.__T_at_min

    @property
    def T_at_max(self):
        """
        float : temperature corresponding to the maximum of
        property correlation in validity range
        """
        return self.__T_at_max

    @property
    @abstractmethod
    def range(self):
        """
        list : temperature validity range for property correlation
        """
        raise NotImplementedError(f"{type(self).__name__}.range "
                                  "NOT IMPLEMENTED")

    @property
    @abstractmethod
    def units(self):
        """
        str : property units
        """
        raise NotImplementedError(f"{type(self).__name__}.units "
                                  "NOT IMPLEMENTED")

    @property
    @abstractmethod
    def long_name(self):
        """
        str : property long name
        """
        raise NotImplementedError(f"{type(self).__name__}.long_name "
                                  "NOT IMPLEMENTED")

    @property
    @abstractmethod
    def description(self):
        """
        str : property description
        """
        raise NotImplementedError(f"{type(self).__name__}.description "
                                  "NOT IMPLEMENTED")



"""__init__ module of lbh15 package"""

__version__ = "1.1.1"
__author__ = "Daniele Panico and Daniele Tomatis"
__company__ = "newcleo"
__date__ = "March 2023"

from .lead import Lead
from .bismuth import Bismuth
from .lbe import LBE
from ._constants import P_ATM
from .properties import lead_properties
from .properties import bismuth_properties
from .properties import lbe_properties



"""Module with the definition of constants for lbh15 package"""
# KEYWORDS
GURVICH_KEYWORD = 'gurvich1991'
SOBOLEV_KEYWORD = 'sobolev2011'

# LEAD CONSTANTS
LEAD_MELTING_TEMPERATURE = 600.6  # [K]
LEAD_MELTING_LATENT_HEAT = 23.07e3  # [J/kg]
LEAD_BOILING_TEMPERATURE = 2021  # [K]
LEAD_VAPORISATION_HEAT = 858.6e3  # [J/kg]

# BISMUTH CONSTANTS
BISMUTH_MELTING_TEMPERATURE = 544.6  # [K]
BISMUTH_MELTING_LATENT_HEAT = 53.3e3  # [J/kg]
BISMUTH_BOILING_TEMPERATURE = 1831  # [K]
BISMUTH_VAPORISATION_HEAT = 856.2e3  # [J/kg]

# LEAD-BISMUTH-EUTECTIC CONSTANTS
LBE_MELTING_TEMPERATURE = 398.0  # [K]
LBE_MELTING_LATENT_HEAT = 38.6e3  # [J/kg]
LBE_BOILING_TEMPERATURE = 1927  # [K]
LBE_VAPORISATION_HEAT = 856.6e3  # [J/kg]

# ATMOSPHERIC PRESSURE
P_ATM = 101325.0  # [Pa]



#!/usr/bin/python3

from setuptools import setup
from setuptools import find_packages
from lbh15 import __version__
from lbh15 import __author__

setup(
    name='lbh15',
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    author=__author__,
    author_email='daniele.panico@newcleo.com, daniele.tomatis@newcleo.com',
    description='Python implementation of liquid metal properties from '
                'Handbook on Lead-bismuth Eutectic Alloy and Lead Properties, '
                'Materials Compatibility, Thermal-hydraulics and Technologies',
    long_description_content_type='text/x-rst',
    long_description=open('README.rst', 'r').read(),
    license='lgpl v3',
    python_requires='>=3.8.10',
    install_requires=['scipy>=1.8.1', 'numpy>=1.22.3'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)

"""
Developers memo for release:
    1. Update package information (on master):
        - Change version in lbh15/__init__.py
        - Change date in lbh15/__init__.py
        - Change date in README.rst

    2. Test the package build and install on test-pypi:
        a. Change the name of the package in setup.py in 'lbh15-test'
        b. python3 setup.py sdist                     # in project directory
        c. twine check dist/*                         # check that package is ok
        d. twine upload --repository testpypi dist/*  # verify upload is successful on test-PyPI
        e. crete a python3 virtual enviroment and verify the correct package installation form test-pypi:
            python3 -m venv venv_install
            source venv_install/bn/activate
            pip3 install wheel
            pip3 install --index-url https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple lbh15-test
            cd tests/
            <execute tests>
        f. If every thing is ok change back the package name in 'lbh15'
        
    3. Create tag (on master):
        a. git tag v<version>
        b. git push <origin_name> <tag_name>
        c. git diff --stat=200 <previous_tag> <tag_name> > log.diff.<tag_name>
        d. You should see the tag on github page as well. Then on github do:
            - https://github.com/newcleo-dev-team/lbh15/releases
            - Click on "Draft a new release"
            - Choose the tag
            - Insert the title name as "Release v<version>"
            - In the description write:
                Changelog between v<previous varsion> and v<version>

                ```
                < content of  log.diff.<tag_name> >
                ```

    4. Upload the package on PyPI:
        a. repeat step 2.b and 2.c
        b. twine upload dist/*             
"""